# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import time

from django import forms
from django.apps import apps
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from django_otp.oath import TOTP
from django_otp.plugins.otp_static.models import StaticDevice, StaticToken
from django_otp.plugins.otp_totp.models import TOTPDevice


TOTP_SESSION_KEY = 'kleides-mfa-totp-key'


class BaseVerifyForm(forms.Form):
    def __init__(
            self, device, plugin, request, unverified_user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.unverified_user = unverified_user
        self.plugin = plugin
        self.device = device
        self.request = request

    def get_user(self):
        if self.is_valid():
            return self.unverified_user
        return None

    def get_device(self):
        if self.is_valid():
            return self.device
        return None


class DeviceVerifyForm(BaseVerifyForm):
    otp_token = forms.CharField(label=_('Token'))

    error_messages = {
        'invalid': _('The token is not valid for this device.'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['otp_token'].widget.attrs.update({
            'autocomplete': 'off', 'autofocus': 'autofocus'})

    def clean(self):
        cleaned_data = super().clean()
        token = cleaned_data.get('otp_token')
        if not token:  # token is a required field.
            return cleaned_data

        # Note that tokens can become invalid once verified.
        if not self.device.verify_token(token):
            raise forms.ValidationError(self.error_messages['invalid'])
        return cleaned_data


class BaseDeviceForm(forms.ModelForm):
    def __init__(self, plugin, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.plugin = plugin
        self.request = request
        if 'name' in self.fields:
            self.fields['name'].initial = self.plugin.name
            self.fields['name'].required = False
            self.fields['name'].widget.attrs['placeholder'] = _(
                'Friendly name')

    def clean_name(self):
        if self.cleaned_data.get('name'):
            return self.cleaned_data['name']
        return self.plugin.name


class DeviceCreateForm(BaseDeviceForm):
    error_messages = {
        'invalid': _('Unable to validate the token with the device.'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # WARNING: The form associates the user with the potential new device
        # instance to allow device functions dependent on the user to be used.
        # The plugin author must take care to only use functions that will not
        # save the instance as this may circumvent the verification process.
        self.instance.user = self.request.user
        self.instance.confirmed = False

    def _post_clean(self):
        super()._post_clean()
        if self.instance.pk is not None:
            raise ValueError(
                'Plugin {} saved the device before completing verification. '
                'The device will be unconfirmed and unusable. '
                'Please review {} {!r}.'.format(
                    self.plugin, self.instance.persistent_id, self.instance))
        self.instance.confirmed = True


class DeviceUpdateForm(BaseDeviceForm):
    pass


class TOTPDeviceCreateForm(DeviceCreateForm):
    otp_token = forms.CharField(label=_('Token'))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['otp_token'].widget.attrs.update({
            'autocomplete': 'off', 'autofocus': 'autofocus'})
        # Store the TOTP key in the session, rotate the key on unbound forms.
        if self.is_bound and TOTP_SESSION_KEY in self.request.session:
            self.instance.key = self.request.session[TOTP_SESSION_KEY]
        else:
            self.request.session[TOTP_SESSION_KEY] = self.instance.key
        # If alternate TOTPDevice settings are needed set them here.
        # self.instance.digits = 8  # increase token length.
        # self.instance.tolerance = 2  # for users that are veeeeery slow.

    def clean(self):
        cleaned_data = super().clean()
        try:
            token = int(cleaned_data.get('otp_token'))
        except (TypeError, ValueError):
            verified = False
        else:
            # django-otp setting.
            OTP_TOTP_SYNC = getattr(settings, 'OTP_TOTP_SYNC', True)
            # Device verification using the current instance.
            totp = TOTP(
                self.instance.bin_key, self.instance.step, self.instance.t0,
                self.instance.digits, self.instance.drift)
            totp.time = time.time()

            verified = totp.verify(
                token, self.instance.tolerance, self.instance.last_t)
            if verified:
                # Device is verified, update attributes and prepare the
                # instance to be saved.
                self.instance.last_t = totp.t()
                if OTP_TOTP_SYNC:
                    self.instance.drift = totp.drift
        if not verified:
            raise forms.ValidationError(self.error_messages['invalid'])
        try:
            return cleaned_data
        finally:
            if TOTP_SESSION_KEY in self.request.session:  # pragma: no cover
                del self.request.session[TOTP_SESSION_KEY]

    class Meta:
        model = TOTPDevice
        fields = ('name', 'otp_token',)


class RecoveryDeviceForm(DeviceUpdateForm):
    token_amount = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.instance = self.request.user.staticdevice_set.get()
        except self.request.user.staticdevice_set.model.DoesNotExist:
            pass

    def save(self, *args, **kwargs):
        instance, created = self.request.user.staticdevice_set.get_or_create(
            defaults={'name': self.plugin.name})
        instance.token_set.all().delete()
        for i in range(self.token_amount):
            instance.token_set.create(token=StaticToken.random_token())
        return instance

    class Meta:
        model = StaticDevice
        fields = ()


if apps.is_installed('otp_yubikey'):  # pragma: no branch
    from otp_yubikey.models import RemoteYubikeyDevice, ValidationService

    class YubikeyDeviceCreateForm(DeviceCreateForm):
        service = forms.ModelChoiceField(
            label=_('Service'), queryset=ValidationService.objects.all(),
            empty_label=None)
        otp_token = forms.CharField(label=_('Token'))

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields['otp_token'].widget.attrs.update({
                'autocomplete': 'off', 'autofocus': 'autofocus'})
            try:
                self.fields['service'].initial = (
                    self.fields['service'].queryset.get().pk)
                self.fields['service'].widget = forms.HiddenInput()
            except ValidationService.MultipleObjectsReturned:
                # Multiple is good, none is bad.
                pass

        def clean(self):
            cleaned_data = super().clean()
            token = self.cleaned_data.get('otp_token')
            service = self.cleaned_data.get('service')
            verified = False
            if token and service:
                self.instance.service = service
                self.instance.public_id = token[:-32]
                verified = self.instance.verify_token(token)
            if not verified:
                raise forms.ValidationError(self.error_messages['invalid'])
            return cleaned_data

        class Meta:
            model = RemoteYubikeyDevice
            fields = ('service', 'name', 'otp_token',)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.contrib.auth import login
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import resolve_url
from django.urls import reverse
from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.http import urlencode

from .mixins import (
    BACKEND_SESSION_KEY, HASH_SESSION_KEY, SESSION_KEY, VERIFIED_SESSION_KEY,
    PluginMixin, UnverifiedUserMixin)
from .. import settings
from ..registry import registry


class LoginView(DjangoLoginView):
    template_name = 'kleides_mfa/login.html'

    def get_success_url(self):
        if not self.request.user.is_verified:
            return resolve_url(settings.SINGLE_FACTOR_URL)
        return super().get_success_url()

    def form_valid(self, form):
        # If the user has any authentication methods remaining they must be
        # used *before the user is logged in*.
        user = form.get_user()
        user_devices = registry.user_devices_with_plugin(user, confirmed=True)
        if user_devices:
            # Store the User data in the session so we can call Django login
            # after verifying a 2nd device.
            session = self.request.session
            session[SESSION_KEY] = user._meta.pk.value_to_string(user)
            session[BACKEND_SESSION_KEY] = user.backend
            if hasattr(user, 'get_session_auth_hash'):
                session[HASH_SESSION_KEY] = user.get_session_auth_hash()

            # Devices are sorted by security/type.
            plugin, device = user_devices[0]
            redirect_url = reverse(
                'kleides_mfa:verify', args=[plugin.slug, device.pk])
            params = urlencode(
                {self.redirect_field_name: self.get_success_url()})
            return HttpResponseRedirect('{}?{}'.format(redirect_url, params))
        # Otherwise the user is authenticated with a single factor and will
        # have to fortify his account by adding authentication methods.
        return super().form_valid(form)


class DeviceVerifyView(UnverifiedUserMixin, PluginMixin, DjangoLoginView):
    template_name_suffix = '_verify_form'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['device'] = self.get_object()
        context['user_devices'] = registry.user_devices_with_plugin(
            self.unverified_user, confirmed=True)
        context['unverified_user'] = self.unverified_user
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['unverified_user'] = self.unverified_user
        kwargs['device'] = self.get_object()
        return kwargs

    def get_form_class(self):
        form_class = self.plugin.get_verify_form_class()
        if form_class is None:
            raise Http404(
                'Plugin {} does not support this'.format(self.plugin))
        return form_class

    def get_object(self):
        # Fetch the the confirmed device of the unverified user.
        return self.plugin.get_user_device(
            self.kwargs['device_id'], self.unverified_user, confirmed=True)

    def form_valid(self, form):
        # User is now verified.
        user = form.get_user()
        # Pass otp device to django-otp.
        user.otp_device = form.get_device()
        # Perform django session login.
        login(self.request, user, self.request.session[BACKEND_SESSION_KEY])
        # Cleanup kleides_mfa session data.
        del self.request.session[SESSION_KEY]
        del self.request.session[BACKEND_SESSION_KEY]
        if hasattr(user, 'get_session_auth_hash'):
            del self.request.session[HASH_SESSION_KEY]
        # Add the last verification time to the session.
        self.request.session[VERIFIED_SESSION_KEY] = force_text(timezone.now())
        return HttpResponseRedirect(self.get_success_url())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from collections import namedtuple

from django.forms import modelform_factory
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _

from .forms import DeviceUpdateForm
from .settings import PLUGIN_PRIORITY

__all__ = ['registry']

KleidesPluginDevices = namedtuple('KleidesPluginDevices', 'plugin devices')
KleidesPluginDevice = namedtuple('KleidesPluginDevice', 'plugin device')


class AlreadyRegistered(Exception):
    pass


class KleidesMfaPlugin():
    def __init__(
            self, name, model, create_form_class=None, update_form_class=None,
            verify_form_class=None, show_create_button=True,
            create_message=None, update_message=None, delete_message=None,
            show_verify_button=True, device_list_javascript=None,
            device_list_template=None):
        self.slug = slugify(name)
        self.name = name
        self.model = model
        self.create_form_class = create_form_class
        self.update_form_class = update_form_class
        self.verify_form_class = verify_form_class
        self.create_message = create_message
        self.update_message = update_message
        self.delete_message = delete_message
        self.device_list_javascript = device_list_javascript
        self.device_list_template = device_list_template
        self.show_create_button = show_create_button
        self.show_verify_button = show_verify_button

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'KleidesMfaPlugin(name={!r}, model={!r})'.format(
            self.name, self.model)

    def get_create_form_class(self):
        return self.create_form_class

    def get_update_form_class(self):
        if self.update_form_class is None:
            return modelform_factory(
                self.model, form=DeviceUpdateForm, fields=('name',))
        return self.update_form_class

    def get_verify_form_class(self):
        return self.verify_form_class

    def get_create_message(self, device):
        if self.create_message is not None:
            message = self.create_message
        else:
            message = _('The {plugin} "{name}" was added successfully.')
        return message.format(plugin=self.name, name=device.name)

    def get_update_message(self, device):
        if self.update_message is not None:
            message = self.update_message
        else:
            message = _('The {plugin} "{name}" was changed successfully.')
        return message.format(plugin=self.name, name=device.name)

    def get_delete_message(self, device):
        if self.delete_message is not None:
            message = self.delete_message
        else:
            message = _('The {plugin} "{name}" was deleted successfully.')
        return message.format(plugin=self.name, name=device.name)

    def get_user_device(self, device_id, user, confirmed=True):
        return self.model.objects.devices_for_user(user, confirmed).get(
            pk=device_id)

    def get_user_devices(self, user, confirmed=True):
        devices = []
        for device in self.model.objects.devices_for_user(user, confirmed):
            devices.append(device)
        return devices


class KleidesMfaPluginRegistry():
    '''
    A registry to store and manage access to KleidesMfaPlugins.
    '''
    def __init__(self):
        self._registry = {}  # plugin.slug -> plugin

    def register_plugin(self, plugin):
        if plugin.slug in self._registry:
            raise AlreadyRegistered(
                'Plugin with slug {} already registered'.format(plugin.slug))
        self._registry[plugin.slug] = plugin

    def register(self, *args, **kwargs):
        self.register_plugin(KleidesMfaPlugin(*args, **kwargs))

    def unregister(self, name_or_slug):
        return self._registry.pop(slugify(name_or_slug))

    def get_plugin(self, slug):
        return self._registry[slug]

    def plugins(self):
        '''
        Return an iterable of registered plugins in settings.PLUGIN_PRIORITY.
        '''
        for slug in PLUGIN_PRIORITY:
            if slug in self._registry:
                yield self._registry[slug]

    def user_has_device(self, user, confirmed=True):
        for plugin in self.plugins():
            if plugin.get_user_devices(user, confirmed):
                return True
        return False

    def plugins_with_user_devices(self, user, confirmed=True):
        '''
        Return an iterable of the plugins with devices registered by the user.
        '''
        return [
            KleidesPluginDevices(
                plugin, plugin.get_user_devices(user, confirmed))
            for plugin in self.plugins()]

    def user_devices_with_plugin(self, user, confirmed=True):
        return [
            KleidesPluginDevice(plugin, device)
            for plugin in self.plugins()
            for device in plugin.get_user_devices(user, confirmed)
        ]


registry = KleidesMfaPluginRegistry()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import functools

from django.utils.functional import SimpleLazyObject

from django_otp import DEVICE_ID_SESSION_KEY
from django_otp.models import Device


class KleidesAuthenticationMiddleware(object):
    """
    This is a replacement for :class:`~django_otp.middleware.OTPMiddleware`
    that uses ``is_anonymous`` instead of ``is_authenticated``. This allows
    us to override the ``is_authenticated`` property to only return True
    when the user has used 2 step authentication.

    This must be installed after
    :class:`~django.contrib.auth.middleware.AuthenticationMiddleware` and
    performs an analogous function. Just as AuthenticationMiddleware populates
    ``request.user`` based on session data, KleidesAuthenticationMiddleware
    populates ``request.user.otp_device`` to the
    :class:`~django_otp.models.Device` object that has verified the user,
    or ``None`` if the user has not been verified.
    """
    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        user = getattr(request, 'user', None)
        if user is not None:
            request.user = SimpleLazyObject(
                functools.partial(self._verify_user, request, user))

        return self.get_response(request)

    def _verify_user(self, request, user):
        """
        Sets OTP-related fields on an authenticated user.
        """
        device = None

        if not user.is_anonymous:
            persistent_id = request.session.get(DEVICE_ID_SESSION_KEY)
            if persistent_id:
                device = Device.from_persistent_id(persistent_id)
                # Ensure the device belongs to the user.
                if device is not None and device.user_id != user.pk:
                    device = None

            if device is None and DEVICE_ID_SESSION_KEY in request.session:
                del request.session[DEVICE_ID_SESSION_KEY]

        user.otp_device = device

        return user



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.conf import settings

# The sorting order of the plugins on the mfa listing using the plugin.slug.
# These are known plugins in order of security.
# hardware tokens > software tokens > backup codes.
PLUGIN_PRIORITY = getattr(settings, 'KLEIDES_MFA_PLUGIN_PRIORITY', [
    'u2f', 'yubikey', 'totp', 'recovery-code',
])

# Patch the AdminSite class and default admin site instance to require
# 2 step authentication.
PATCH_ADMIN = getattr(settings, 'KLEIDES_MFA_PATCH_ADMIN', True)
# Patch the active User model and the AnonymousUser with 2 step authentication
# properties.
PATCH_USER = getattr(settings, 'KLEIDES_MFA_PATCH_USER', True)

# The default url to redirect to after login when the user has no 2 step
# authentication devices active on the account.
SINGLE_FACTOR_URL = getattr(
    settings, 'KLEIDES_MFA_SINGLE_FACTOR_URL', 'kleides_mfa:index')



# -*- coding: utf-8 -*-

"""Kleides Multi Factor Authentication."""

__author__ = """Harm Geerts"""
__email__ = 'hgeerts@osso.nl'
__version__ = '0.1.4'

default_app_config = 'kleides_mfa.apps.KleidesMfaConfig'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.urls import path

from . import views

app_name = 'kleides_mfa'

urlpatterns = [
    path('login/', views.LoginView.as_view(), name='login'),
    path('list/', views.DeviceListView.as_view(), name='index'),
    path(
        '<slug:plugin>/create/', views.DeviceCreateView.as_view(),
        name='create'),
    path(
        '<slug:plugin>/update/<path:device_id>/',
        views.DeviceUpdateView.as_view(), name='update'),
    path(
        '<slug:plugin>/verify/<path:device_id>/',
        views.DeviceVerifyView.as_view(), name='verify'),
    path(
        '<slug:plugin>/delete/<path:device_id>/',
        views.DeviceDeleteView.as_view(), name='delete'),
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# kleides-mfa test project
from os.path import abspath, dirname, join


def project_path(path):
    return abspath(join(dirname(__file__), path))


DEBUG = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'django.contrib.messages',

    'django_otp',
    'django_otp.plugins.otp_email',
    'django_otp.plugins.otp_hotp',
    'django_otp.plugins.otp_static',
    'django_otp.plugins.otp_totp',
    'otp_yubikey',

    'kleides_mfa',
    'crispy_forms',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'kleides_mfa.middleware.KleidesAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

AUTHENTICATION_BACKENDS = [
    'django.contrib.auth.backends.ModelBackend',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'DIRS': [
            project_path('templates'),
        ],
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

SECRET_KEY = 'XXX'

ROOT_URLCONF = 'tests.urls'

CRISPY_TEMPLATE_PACK = 'bootstrap4'

OTP_TOTP_ISSUER = 'Kleides MFA'

ALLOWED_HOSTS = [
    'localhost',
    '127.0.0.1',
    'testserver',
]
STATIC_URL = '/static/'
STATIC_ROOT = project_path('../static/')

LOGIN_URL = 'kleides_mfa:login'
LOGIN_REDIRECT_URL = 'kleides_mfa:index'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.contrib import messages
from django.http import Http404
from django.views.generic import (
    CreateView, DeleteView, TemplateView, UpdateView)

from django_otp import DEVICE_ID_SESSION_KEY, login as django_otp_login

from ..registry import registry
from .mixins import (
    MultiFactorRequiredMixin, PluginMixin, SetupOrMFARequiredMixin,
    SingleFactorRequiredMixin)


class DeviceListView(SingleFactorRequiredMixin, TemplateView):
    template_name = 'kleides_mfa/plugin_list.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['plugins'] = registry.plugins_with_user_devices(
            self.request.user, confirmed=None)
        return context


class DeviceCreateView(SetupOrMFARequiredMixin, PluginMixin, CreateView):
    # XXX notification via email ?
    template_name_suffix = '_create_form'

    def get_form_class(self):
        form_class = self.plugin.get_create_form_class()
        if form_class is None:
            raise Http404(
                'Plugin {} does not support this'.format(self.plugin))
        return form_class

    def form_valid(self, form):
        response = super().form_valid(form)
        # This is the users first device, use it to verify the user.
        if not self.request.user.is_verified:
            django_otp_login(self.request, self.object)
        messages.success(
            self.request, self.plugin.get_create_message(self.object))
        return response


class DeviceUpdateView(MultiFactorRequiredMixin, PluginMixin, UpdateView):
    def get_form_class(self):
        return self.plugin.get_update_form_class()

    def form_valid(self, form):
        messages.success(
            self.request, self.plugin.get_update_message(self.object))
        return super().form_valid(form)


class DeviceDeleteView(MultiFactorRequiredMixin, PluginMixin, DeleteView):
    # XXX notification via email ?
    # XXX password confirmation ?
    def delete(self, request, *args, **kwargs):
        message = self.plugin.get_delete_message(self.get_object())
        response = super().delete(request, *args, **kwargs)
        messages.warning(request, message)
        # User has removed all authentication methods, disable his access.
        if not registry.user_has_device(self.request.user, confirmed=True):
            self.request.user.otp_device = None
            if DEVICE_ID_SESSION_KEY in self.request.session:
                del self.request.session[DEVICE_ID_SESSION_KEY]

        return response



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from base64 import b32decode
from urllib.parse import parse_qs, urlsplit

from django.test import TestCase
from django.test.utils import override_settings

from django_otp.oath import TOTP

from .factories import UserFactory


class DjangoOtpTotpTestCase(TestCase):
    def totp_from_form(self, form):
        url = urlsplit(form.instance.config_url)
        params = parse_qs(url.query)
        return TOTP(
            b32decode(params['secret'][0]),
            step=int(params['period'][0]),
            digits=int(params['digits'][0]), t0=0, drift=0)

    def login(self, user, redirect_to='/list/'):
        response = self.client.post(
            '/login/',
            {'username': user.username, 'password': user.raw_password},
            follow=True)
        self.assertRedirects(response, redirect_to)
        return response

    @override_settings(OTP_TOTP_SYNC=True, OTP_TOTP_THROTTLE_FACTOR=0)
    def test_totp(self):
        user = UserFactory()
        self.login(user)

        response = self.client.post(
            '/totp/create/', {'otp_token': 'XXX', 'name': 'My Phone'})
        self.assertContains(
            response, 'Unable to validate the token with the device.')
        response = self.client.post(
            '/totp/create/', {'otp_token': '123456', 'name': 'My Phone'})
        self.assertContains(
            response, 'Unable to validate the token with the device.')

        # User takes the TOTP config and applies it to his device.
        response = self.client.get('/totp/create/')
        totp = self.totp_from_form(response.context['form'])

        # Confirm the user can generate tokens and the device is configured.
        registration_token = totp.token()
        response = self.client.post(
            '/totp/create/', {
                'otp_token': registration_token, 'name': 'My Phone'})
        self.assertRedirects(response, '/list/')
        device = user.totpdevice_set.get()
        self.assertTrue(device.confirmed)
        self.assertEqual(device.name, 'My Phone')

        response = self.client.post(
            '/totp/update/{}/'.format(device.pk), {'name': 'Acme ID'})
        self.assertRedirects(response, '/list/')
        device = user.totpdevice_set.get()
        self.assertEqual(device.name, 'Acme ID')

        self.client.logout()

        # Login without TOTP.
        verify_url = '/totp/verify/{}/'.format(device.pk)
        self.login(user, '{}?next=/list/'.format(verify_url))

        # User is forced to use 2 step authentication.
        response = self.client.get('/list/')
        self.assertRedirects(response, '/login/?next=/list/')
        response = self.client.post('/totp/delete/{}/'.format(device.pk))
        self.assertRedirects(
            response, '/login/?next=/totp/delete/{}/'.format(device.pk))

        # Token reuse is not allowed so the registration is no longer valid.
        response = self.client.post(
            verify_url, {'otp_token': registration_token})
        self.assertContains(
            response, 'The token is not valid for this device.')

        # Default tolerance is 1, increase drift to force next token.
        totp.drift = 1
        response = self.client.post(verify_url, {'otp_token': totp.token()})
        self.assertRedirects(response, '/list/')

        response = self.client.post(
            '/totp/delete/{}/'.format(device.pk), follow=True)
        self.assertContains(
            response, 'The TOTP &quot;Acme ID&quot; was deleted successfully.')

    @override_settings(OTP_TOTP_SYNC=False)
    def test_otp_sync(self):
        user = UserFactory()
        self.login(user)

        response = self.client.get('/totp/create/')
        totp = self.totp_from_form(response.context['form'])
        totp.drift = 1
        response = self.client.post(
            '/totp/create/', {'otp_token': totp.token()})
        self.assertRedirects(response, '/list/')
        device = user.totpdevice_set.get()
        self.assertTrue(device.confirmed)
        self.assertEqual(device.drift, 0)



from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import include, path
from django.views.generic.base import RedirectView

urlpatterns = [
    path('', RedirectView.as_view(url='/list/')),
    path('', include('kleides_mfa.urls')),
    path('admin/', admin.site.urls),
] + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from .auth import LoginView, DeviceVerifyView
from .devices import (
    DeviceDeleteView, DeviceCreateView, DeviceListView, DeviceUpdateView)

__all__ = [
    'DeviceDeleteView', 'DeviceCreateView', 'DeviceListView',
    'DeviceUpdateView', 'DeviceVerifyView', 'LoginView', 'VerifyView',
]
